"""
Module for new-style block parsing. Make sure all functions are up-to-date and used.
"""


def f2(lines, tag, i=0, j=0):
    for k in range(i, len(lines)):
        index = lines[k].find(tag, j if k == i else 0)
        if index >= 0:
            return k, index
    return None, None


def block_iterate(lines, tag):
    contents = {'joined': lines}
    while True:
        contents = block_split(contents['joined'], tag)
        if contents is None:
            break

        yield  contents

def block_join(contents):
    blk = contents['block']
    if len(blk) == 1:
        blk = [blk[0] + contents['post1'] + " " + contents['post2']]
    elif len(blk) > 1:
        blk = [blk[0] + contents['post1']] + blk[1:-1] + [blk[-1] + contents['post2']]
    return contents['first'] + blk + contents['last']


def block_split(lines, tag):
    stag = tag[:2]  # Start of any next tag.


    contents = {}
    i, j = f2(lines, tag)

    def get_tag_args(line):
        # line = line.strip()
        k = line.find(" ")
        tag_args = ((line[:k + 1] if k >= 0 else line)[len(tag):] ).strip().split(";")
        tag_args = [t.strip() for t in tag_args]
        # if len(tag_args) == 0:
        #     return {'': ''}  # No name.
        tag_args = [t for t in tag_args if len(t) > 0]
        tag_args = dict([t.split("=") if "=" in t else (t.lower().strip(), True) for t in tag_args])

        if '' not in tag_args:
            tag_args[''] = ''

        return tag_args

    if i is None:
        return None
    else:
        # print( lines[i] )

        start_tag_args = get_tag_args(lines[i][j:])
        START_TAG = f"{tag}={start_tag_args['']}" if start_tag_args[''] != '' else tag
        END_TAG = START_TAG
        i2, j2 = f2(lines, END_TAG, i=i, j=j+1)
        if i2 == None:
            END_TAG = tag
            i2, j2 = f2(lines, END_TAG, i=i, j=j+1)
        if i2 == None:
            print("\n".join( lines[i:]))
            raise Exception("Did not find matching tag", tag)

    if i == i2:
        # Splitting a single line. To reduce confusion, this will be treated slightly differently:
        l2 = lines[:i] + [lines[i][:j2], lines[i][j2:]] + lines[i2+1:]
        c2 = block_split(l2, tag=tag)
        c2['block'].pop()
        c2['joined'] = block_join(c2)
        return c2
    else:
        contents['first'] = lines[:i]
        contents['last'] = lines[i2+1:]

        def argpost(line, j):
            nx_tag = line.find(stag, j+1)
            arg1 = line[j+len(tag):nx_tag]
            if nx_tag >= 0:
                post = line[nx_tag:]
            else:
                post = ''
            return arg1, post

        contents['arg1'], contents['post1'] = argpost(lines[i], j)
        contents['arg2'], contents['post2'] = argpost(lines[i2], j2)
        blk = [lines[i][:j]] + lines[i+1:i2] + [lines[i2][:j2]]
        contents['block'] = blk
        contents['joined'] = block_join(contents)
        contents['start_tag_args'] = start_tag_args
        contents['name'] = start_tag_args['']
        return contents


def indent(l):
    return l[:len(l) - len(l.lstrip())]


def full_strip(lines, tags=None):
    if tags is None:
        tags = ["#!s", "#!o", "#!f", "#!b"]
    for t in tags:
        lines = strip_tag(lines, t)
    return lines


def strip_tag(lines, tag):
    lines2 = []
    for l in lines:
        dx = l.find(tag)
        if dx > 0:
            l = l[:dx]
            if len(l.strip()) == 0:
                l = None
        if l is not None:
            lines2.append(l)
    return lines2


import os
import io
from pybtex import plugin
from pybtex.database.input import bibtex
from warnings import warn

### Newstyle loading.
def get_aux(auxfile):
    # paths = get_paths()
    # auxfile = os.path.join(paths['02450public'], auxfile)
    if not os.path.exists(auxfile):
        # print(auxfile)

        warn("Could not find bibtex file: "+ auxfile)
        return {}

    with open(auxfile, 'r') as f:
        items = f.readlines()
    entries = {}
    for e in items:
        e = e.strip()
        if e.startswith("\\newlabel") and "@cref" in e:
            # print(e)
            i0 = e.find("{")
            i1 = e.find("@cref}")
            key = e[i0+1:i1]

            j0 = e.find("{{[", i0)+3
            j1 = e.find("}", j0)

            val = e[j0:j1]

            label = val[:val.find("]")]
            number = val[val.rfind("]")+1:]

            if label == "equation":
                nlabel = f"eq. ({number})"
            else:
                nlabel = label.capitalize() + " " + number

            # coderef = "\\cite[%s]{%s}"%(nlabel, bibtex) if bibtex is not None else None

            entries[key] = {'nicelabel': nlabel, 'rawlabel': label, 'number': number}
    return entries


def get_bibtex(bibfile):
    """
    all references.
    """
    if not os.path.exists(bibfile):
        return None

    pybtex_style = plugin.find_plugin('pybtex.style.formatting', 'alpha')()
    pybtex_html_backend = plugin.find_plugin('pybtex.backends', 'html')()
    pybtex_plain_backend = plugin.find_plugin('pybtex.backends', 'plaintext')()
    pybtex_parser = bibtex.Parser()

    with open(bibfile, 'r', encoding='utf8') as f:
        data = pybtex_parser.parse_stream(f)

    data_formatted = pybtex_style.format_entries(data.entries.values())
    refs = {}

    # if 'auxfile' in gi:
    #     all_references = parse_aux(gi['auxfile'], bibtex=gi['bibtex'])
    # else:
    #     all_references = {}

    for entry in data_formatted:
        output = io.StringIO()
        output_plain = io.StringIO()
        pybtex_plain_backend.output = output_plain.write
        pybtex_html_backend.output = output.write
        pybtex_html_backend.write_entry(entry.key, entry.label, entry.text.render(pybtex_html_backend))

        pybtex_plain_backend.write_entry(entry.key, entry.label, entry.text.render(pybtex_plain_backend))

        html = output.getvalue()
        plain = output_plain.getvalue()

        entry.text.parts[-2].__str__()
        url = ""
        for i,p in enumerate(entry.text.parts):
            if "\\url" in p.__str__():
                url = entry.text.parts[i+1]
                break
        url = url.__str__()
        i1 = html.find("\\textbf")
        i2 = html.find("</span>", i1)
        dht = html[i1:i2]
        dht = dht[dht.find(">")+1:]
        html = html[:i1] + " <b>"+dht+"</b> " + html[i2+7:]

        plain = plain.replace("\\textbf ", "")
        iu = plain.find("URL")
        if iu > 0:
            plain = plain[:iu]

        refs[entry.key] = {'html': html,
                           'plain': plain,
                           'label': entry.label,
                           'filename': url,
                           }

    return refs


def find_tex_cite(s, start=0, key="\\cite"):
    txt = None
    i = s.find(key, start)
    if i < 0:
        return (i,None), None, None
    j = s.find("}", i)
    cite = s[i:j + 1]

    if cite.find("[") > 0:
        txt = cite[cite.find("[") + 1:cite.find("]")]

    reference = cite[cite.find("{") + 1:cite.find("}")]
    return (i, j), reference, txt

### Oldstyle loading
# def get_references(bibfile, gi):
#     """
#     all references.
#     """
#     if not os.path.exists(bibfile):
#         return None
#
#     pybtex_style = plugin.find_plugin('pybtex.style.formatting', 'alpha')()
#     pybtex_html_backend = plugin.find_plugin('pybtex.backends', 'html')()
#     pybtex_plain_backend = plugin.find_plugin('pybtex.backends', 'plaintext')()
#     pybtex_parser = bibtex.Parser()
#
#     with open(bibfile, 'r', encoding='utf8') as f:
#         data = pybtex_parser.parse_stream(f)
#
#     data_formatted = pybtex_style.format_entries(data.entries.values())
#     refs = {}
#
#     if 'auxfile' in gi:
#         all_references = parse_aux(gi['auxfile'], bibtex=gi['bibtex'])
#     else:
#         all_references = {}
#
#     for entry in data_formatted:
#         output = io.StringIO()
#         output_plain = io.StringIO()
#         pybtex_plain_backend.output = output_plain.write
#         pybtex_html_backend.output = output.write
#         pybtex_html_backend.write_entry(entry.key, entry.label, entry.text.render(pybtex_html_backend))
#
#         pybtex_plain_backend.write_entry(entry.key, entry.label, entry.text.render(pybtex_plain_backend))
#
#         html = output.getvalue()
#         plain = output_plain.getvalue()
#
#         entry.text.parts[-2].__str__()
#         url = ""
#         for i,p in enumerate(entry.text.parts):
#             if "\\url" in p.__str__():
#                 url = entry.text.parts[i+1]
#                 break
#         url = url.__str__()
#         i1 = html.find("\\textbf")
#         i2 = html.find("</span>", i1)
#         dht = html[i1:i2]
#         dht = dht[dht.find(">")+1:]
#         html = html[:i1] + " <b>"+dht+"</b> " + html[i2+7:]
#
#         plain = plain.replace("\\textbf ", "")
#         iu = plain.find("URL")
#         if iu > 0:
#             plain = plain[:iu]
#
#         refs[entry.key] = {'html': html,
#                            'plain': plain,
#                            'label': entry.label,
#                            'filename': url,
#                            'references': all_references}
#
#     newref = {}
#     ls = lambda x: x if isinstance(x, list) else [x]
#     if 'tex_command' in gi:
#         for cmd, aux, display in zip( ls(gi['tex_command']), ls(gi['tex_aux'] ), ls( gi['tex_display'] ) ):
#             ax = parse_aux(aux, bibtex=gi['bibtex'])
#             for k in ax:
#                 ax[k]['pyref'] = display%(ax[k]['nicelabel'],)
#             newref[cmd] = ax
#
#     return refs, newref
#
#
# def parse_aux(auxfile, bibtex=None):
#     # paths = get_paths()
#     paths = {}
#     auxfile = os.path.join(paths['02450public'], auxfile)
#     if not os.path.exists(auxfile):
#         print(auxfile)
#         from warnings import warn
#         warn("Could not find file")
#         return {}
#
#     with open(auxfile, 'r') as f:
#         items = f.readlines()
#     entries = {}
#     for e in items:
#         e = e.strip()
#         if e.startswith("\\newlabel") and "@cref" in e:
#             # print(e)
#             i0 = e.find("{")
#             i1 = e.find("@cref}")
#             key = e[i0+1:i1]
#
#             j0 = e.find("{{[", i0)+3
#             j1 = e.find("}", j0)
#
#             val = e[j0:j1]
#
#             label = val[:val.find("]")]
#             number = val[val.rfind("]")+1:]
#
#             if label == "equation":
#                 nlabel = f"eq. ({number})"
#             else:
#                 nlabel = label.capitalize() + " " + number
#
#             coderef = "\\cite[%s]{%s}"%(nlabel, bibtex) if bibtex is not None else None
#             entries[key] = {'pyref': coderef, 'nicelabel': nlabel, 'rawlabel': label, 'number': number}
#     return entries




import os
import re

from snipper.block_parsing import full_strip
from snipper.fix_i import run_i
from snipper.fix_r import fix_r
from snipper.fix_s import save_s
from snipper.fix_cite import fix_citations
from snipper.fix_bf import fix_f, fix_b
from snipper.fix_o import run_o


def rem_nonprintable_ctrl_chars(txt):
    """Remove non_printable ascii control characters """
    try:
        txt = re.sub(r'[^\x20-\x7E|\x09-\x0A]','', txt)
        # remove non-ascii characters
        txt = repr(txt).decode('unicode_escape').encode('ascii','ignore')[1:-1]
    except Exception as exception:
        print(exception)
    return txt


def censor_code(lines, keep=True):
    dbug = True
    lines = fix_f(lines, dbug)
    lines, nB, cut = fix_b(lines, keep=True)
    return lines



def censor_file(file, run_files=True, run_out_dirs=None, cut_files=True, solution_list=None,
                censor_files=True,
                base_path=None,
                strict=True,
                references=None,
                license_head=None):

    if references == None:
        references = {}

    dbug = False
    with open(file, 'r', encoding='utf8') as f:
        s = f.read()
        s = s.lstrip()
        lines = s.split("\n")
        for k, l in enumerate(lines):
            if l.find(" # !") > 0:
                print(f"{file}:{k}> bad snipper tag, fixing")
            lines[k] = l.replace("# !", "#!")

        try:
            lines = fix_citations(lines, references, strict=strict)
        except IndexError as e:
            print(e)
            print("Error in file, cite/reference tag not found!>", file)
            raise e

        if (run_files or cut_files) and run_out_dirs is not None:
            ofiles = []
            for rod in [run_out_dirs]:
                ofiles.append(os.path.join(rod, os.path.basename(file).split(".")[0]) )
            ofiles[0] = ofiles[0].replace("\\", "/")

            if run_files:
                run_o(lines, file=file, output=ofiles[0])
                run_i(lines, file=file, output=ofiles[0])
            if cut_files:
                save_s(lines, file_path=os.path.relpath(file, base_path), output_dir=run_out_dirs)  # save file snips to disk
        lines = full_strip(lines, ["#!s", "#!o", '#!i'])

        if censor_files:
            lines = fix_f(lines, dbug)
            lines, nB, cut = fix_b(lines)
        else:
            nB = 0
        lines = fix_r(lines)

        if censor_files and len(cut) > 0 and solution_list is not None:
            fname = file.__str__()
            i = fname.find("irlc")
            # wk = fname[i+5:fname.find("\\", i+6)]
            # sp = paths['02450students'] +"/solutions/"
            # if not os.path.exists(sp):
            #     os.mkdir(sp)
            # sp = sp + wk
            # if not os.path.exists(sp):
            #     os.mkdir(sp)
            sols = []
            stext = ["\n".join(lines) for lines in cut]
            for i,sol in enumerate(stext):
                sols.append( (sol,) )
                # sout = sp + f"/{os.path.basename(fname)[:-3]}_TODO_{i+1}.py"
                # wsol = any([True for s in solution_list if os.path.basename(sout).startswith(s)])
                # print(sout, "(published)" if wsol else "")
                # if wsol:
                #     with open(sout, "w") as f:
                #         f.write(sol)

        if len(lines) > 0 and len(lines[-1])>0:
            lines.append("")
        s2 = "\n".join(lines)

    if license_head is not None:
        s2 = fix_copyright(s2, license_head)


    with open(file, 'w', encoding='utf-8') as f:
        f.write(s2)
    return nB


def fix_copyright(s, license_head):
    return "\n".join( ["# " + l.strip() for l in license_head.splitlines()] ) +"\n" + s

# lines: 294, 399, 420, 116


from snipper.load_citations import find_tex_cite
from snipper.legacy import COMMENT


def fix_citations(lines, references, strict=True):
    lines = fix_aux(lines, aux=references.get('aux', {}) )
    for cm in references.get('commands', []):
        lines = fix_aux_special(lines, aux=cm['aux'], command=cm['command'], output=cm['output'])

    lines = fix_bibtex(lines, bibtex=references.get('bibtex', {}))
    return lines


def fix_aux_special(lines, aux, command='\\nref', output='\cite[%s]{my_bibtex_entry}'):
    # out = output %
    daux = {name: {'nicelabel': output % v['nicelabel'] } for name, v in aux.items()}
    l2 = fix_single_reference(lines, aux=daux, cmd=command, strict=True)
    return l2

def fix_aux(lines, aux, strict=True):
    l2 = fix_single_reference(lines, aux=aux, cmd="\\ref", strict=True)
    # print("\n".join(l2))
    return l2

def fix_bibtex(lines, bibtex):
    s = "\n".join(lines)
    i = 0
    all_refs = []
    while True:
        (i, j), reference, txt = find_tex_cite(s, start=i, key="\\cite")
        if i < 0:
            break
        if reference not in bibtex:
            raise IndexError("no such reference: " + reference)
        ref = bibtex[reference]
        label = ref['label']
        rtxt = f"({label}" + (", "+txt if txt is not None else "") + ")"
        r = ref['plain']
        if r not in all_refs:
            all_refs.append(r)
        s = s[:i] + rtxt + s[j+1:]
        i = i + len(rtxt)

    if len(all_refs) > 0:
        if not s.startswith(COMMENT):
            s = f"{COMMENT}\n{COMMENT}\n" + s
        i = s.find(COMMENT, s.find(COMMENT)+1)
        all_refs = ["  " + r.strip() for r in all_refs]
        s = s[:i] + "References:\n" + "\n".join(all_refs) +"\n"+ s[i:]

    # s = s.replace(cpr, info['code_copyright'])
    return s.splitlines()

# def fix_references(lines, info, strict=True):
#     assert False
#     for cmd in info['new_references']:
#         lines = fix_single_reference(lines, cmd, info['new_references'][cmd], strict=strict)
#     return lines

def fix_single_reference(lines, cmd, aux, strict=True):
    references = aux
    s = "\n".join(lines)
    i = 0
    while True:
        (i, j), reference, txt = find_tex_cite(s, start=i, key=cmd)
        if i < 0:
            break
        if reference not in references:
            er = "cref label not found for label: " + reference
            if strict:
                raise IndexError(er)
            else:
                print(er)
                continue
        r = references[reference]
        rtxt = r['nicelabel']
        s = s[:i] + rtxt + s[j + 1:]
        i = i + len(rtxt)
        print(cmd, rtxt)

    lines = s.splitlines(keepends=False)
    return lines


import functools
import os
from snipper.legacy import block_process
from snipper.block_parsing import indent


def run_o(lines, file, output):
    def block_fun(lines, start_extra, end_extra, art, output, **kwargs):
        id = indent(lines[0])
        outf = output + ("_" + art if art is not None else "") + ".txt"
        l2 = []
        l2 += [id + "import sys", id + f"sys.stdout = open('{outf}', 'w')"]
        l2 += lines
        l2 += [indent(lines[-1]) + "sys.stdout = sys.__stdout__"]
        return l2, None
    try:
        lines2, didfind, extra, _ = block_process(lines, tag="#!o", block_fun=functools.partial(block_fun, output=output) )
    except Exception as e:
        print("Bad file: ", file)
        print("I was cutting the #!o tag")
        print("\n".join( lines) )
        raise(e)

    if didfind:
        fp, ex = os.path.splitext(file)
        file_run = fp + "_RUN_OUTPUT_CAPTURE" +ex
        if os.path.exists(file_run):
            print("file found mumble...")
        else:
            with open(file_run, 'w', encoding="utf-8") as f:
                f.write("\n".join(lines2) )
            cmd = "python " + file_run
            import subprocess
            s = subprocess.check_output(cmd, shell=True)
            # s,ok = execute_command(cmd.split(), shell=True)
            print(s)
            os.remove(file_run)


import functools
from snipper.legacy import gcoms
from snipper.block_parsing import indent
from snipper.block_parsing import block_split, block_join


def fix_f(lines, debug):
    lines2 = []
    i = 0
    while i < len(lines):
        l = lines[i]
        dx = l.find("#!f")
        if dx >= 0:
            l_head = l[dx+3:].strip()
            l = l[:dx]
            lines2.append(l)
            id = indent(lines[i+1])
            for j in range(i+1, 10000):
                jid = len( indent(lines[j]) )
                if  j+1 == len(lines) or ( jid < len(id) and len(lines[j].strip() ) > 0):
                    break

            if len(lines[j-1].strip()) == 0:
                j = j - 1
            funbody = "\n".join( lines[i+1:j] )
            if i == j:
                raise Exception("Empty function body")
            i = j
            comments, funrem = gcoms(funbody)
            comments = [id + c for c in comments]
            if len(comments) > 0:
                lines2 += comments[0].split("\n")
            f = [id + l.strip() for l in funrem.splitlines()]
            f[0] = f[0] + "#!b"
            errm = l_head if len(l_head) > 0 else "Implement function body"
            f[-1] = f[-1] + f' #!b {errm}'
            lines2 += f
        else:
            lines2.append(l)
            i += 1
    return lines2

# stats = {'n': 0}
def _block_fun(lines, start_extra, end_extra, keep=False, silent=False):
    id = indent(lines[0])
    lines = lines[1:] if len(lines[0].strip()) == 0 else lines
    lines = lines[:-1] if len(lines[-1].strip()) == 0 else lines
    cc = len(lines)
    ee = end_extra.strip()
    if len(ee) >= 2 and ee[0] == '"':
        ee = ee[1:-1]
    start_extra = start_extra.strip()
    if keep:
        l2 = ['GARBAGE'] * cc
    else:
        if silent:
            l2 = []
            cc = 0
        else:
            l2 = ([id + start_extra] if len(start_extra) > 0 else []) + [id + f"# TODO: {cc} lines missing.",
                                                                         id + f'raise NotImplementedError("{ee}")']
    return l2, cc

def fix_b(lines, keep=False):
    cutout = []
    n = 0
    while True:
        b = block_split(lines, tag="#!b")
        if b == None:
            break
        args = {k:v for k, v in b['start_tag_args'].items() if len(k) > 0}
        cutout += b['block']
        b['block'], dn = _block_fun(b['block'], start_extra=b['arg1'], end_extra=b['arg2'], **args, keep=keep)
        lines = block_join(b)
        n += dn

    # lines2, _, _, cutout = block_process(lines, tag="#!b", block_fun=functools.partial(block_fun, stats=stats))
    return lines, n, cutout


from snipper.block_parsing import indent


def fix_r(lines):
    for i,l in enumerate(lines):
        if "#!r" in l:
            lines[i] = indent(l) + l[l.find("#!r") + 3:].lstrip()
    return lines



import os, shutil
from snipper.snipper_main import censor_file
from pathlib import Path
import time
import fnmatch
import tempfile

def snip_dir(source_dir,  # Sources
             dest_dir=None,  # Will write to this directory
             output_dir=None, # Where snippets are going to be stored
             references=None, # Reference database
             exclude=None, clean_destination_dir=True,
             run_files=True,  # Run #!o tags and #!i tags
             cut_files=True,   # censor files.
             license_head=None,
             ):

    if dest_dir == None:
        dest_dir = tempfile.mkdtemp()
        print("[snipper]", "no destination dir was specified so using nonsense destination:", dest_dir)

    if references == None:
        references = dict(aux=None, bibtex=None,  commands=[])

    if exclude == None:
        exclude = []

    exclude += ["*__pycache__*"]  # Just...no.
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    source_dir = os.path.abspath(source_dir)
    dest_dir = os.path.abspath(dest_dir)
    if output_dir == None:
        output_dir = os.path.dirname(source_dir) + "/output"

    output_dir = os.path.abspath(output_dir)
    if os.path.samefile( source_dir, dest_dir):
        raise Exception("Source and destination is the same")

    if clean_destination_dir:
        shutil.rmtree(dest_dir)
    os.makedirs(dest_dir)

    out = dest_dir
    hw = {'base': source_dir}

    print(f"[snipper] Synchronizing directories: {hw['base']} -> {out}")
    if os.path.exists(dest_dir):
        shutil.rmtree(dest_dir)

    shutil.copytree(source_dir, dest_dir)
    time.sleep(0.1)

    ls = list(Path(dest_dir).glob('**/*.*'))
    acceptable = []
    for l in ls:
        split = os.path.normpath(os.path.relpath(l, dest_dir))
        m = [fnmatch.fnmatch(split, ex) for ex in exclude]
        acceptable.append( (l, not any(m) ))

    # for f,ac in acceptable:
    #     if not ac:
    #         print(f)

    # print(acceptable)
    # now we have acceptable files in list.
    # run_out_dirs = ["./output"]
    n = 0
    # edirs = [os.path.join(out, f_) for f_ in hw['exclusion']]  # Exclude directories on exclude list (speed)
    # edirs = {os.path.normpath(os.path.dirname(f_) if not os.path.isdir(f_) else f_) for f_ in edirs}
    # edirs.remove(os.path.normpath(out))
    for f, accept in acceptable:
        if os.path.isdir(f) or not str(f).endswith(".py") or  str(f).endswith("_grade.py"): # We only touch .py files.
            continue
        # f_dir = os.path.normpath(f if os.path.isdir(f) else os.path.dirname(f))
        if accept:
            # slist = hw['solutions'] if not CE else [os.path.basename(f)[:-3]]
            # base = None
            # if students_irlc_tools is not None:
            #     base = os.path.relpath(str(f), students_irlc_tools + "/..")
            #     base = base.replace("\\", "/")

            # if "assignments" in str(f) and "_grade.py" in str(f):
            #     continue
            # info = {'new_references': [], 'code_copyright': 'Example student code. This file is automatically generated from the files in the instructor-directory'}
            # paths = {}
            solution_list = []
            kwargs = {}
            # cut_files = True
            # copyright()

            # run_files = True
            nrem = censor_file(f, run_files=run_files, run_out_dirs=output_dir, cut_files=cut_files, solution_list=solution_list,
                               base_path=dest_dir,
                               references=references,
                               license_head=license_head,
                               **kwargs)
            if nrem > 0:
                print(f"{nrem}> {f}")
            n += nrem
    for rm_file, accept in acceptable:
        rm_file = os.path.abspath(rm_file)
        if not accept:
            if os.path.isfile(rm_file):
                os.remove(rm_file)
            else:
                if os.path.isdir(rm_file + "\\"):
                    shutil.rmtree(rm_file)

    return n



COMMENT = '"""'


def gcoms(s):
    coms = []
    while True:
        i = s.find(COMMENT)
        if i >= 0:
            j = s.find(COMMENT, i+len(COMMENT))+3
        else:
            break
        if j < 0:
            raise Exception("comment tag not closed")
        coms.append(s[i:j])
        s = s[:i] + s[j:]
        if len(coms) > 10:
            print("long comments in file", i)
    return coms, s


def block_process(lines, tag, block_fun):
    i = 0
    didfind = False
    lines2 = []
    block_out = []
    cutout = []
    while i < len(lines):
        l = lines[i]
        dx = l.find(tag)
        if dx >= 0:
            if l.find(tag, dx+1) > 0:
                j = i
            else:
                for j in range(i + 1, 10000):
                    if j >= len(lines):
                        print("\n".join(lines))
                        print("very bad end-line j while fixing tag", tag)
                        raise Exception("Bad line while fixing", tag)
                    if lines[j].find(tag) >= 0:
                        break

            pbody = lines[i:j+1]
            if i == j:
                start_extra = lines[j][dx:lines[j].rfind(tag)].strip()
                end_extra = lines[j][lines[j].rfind(tag) + len(tag):].strip()
            else:
                start_extra = lines[i][dx:].strip()
                end_extra = lines[j][lines[j].rfind(tag) + len(tag):].strip()

            cutout.append(pbody)
            tmp_ = start_extra.split("=")
            arg = None if len(tmp_) <= 1 else tmp_[1].split(" ")[0]
            start_extra = ' '.join(start_extra.split(" ")[1:] )

            pbody[0] = pbody[0][:dx]
            if j > i:
                pbody[-1] = pbody[-1][:pbody[-1].find(tag)]

            nlines, extra = block_fun(lines=pbody, start_extra=start_extra, end_extra=end_extra, art=arg, head=lines[:i], tail=lines[j+1:])
            lines2 += nlines
            block_out.append(extra)
            i = j+1
            didfind = True
            if "!b" in end_extra:
                assert(False)
        else:
            lines2.append(l)
            i += 1

    return lines2, didfind, block_out, cutout


import functools
import textwrap
from snipper.legacy import block_process
from snipper.block_parsing import full_strip
import os
if os.name == 'nt':
    import wexpect as we
else:
    import pexpect as we


def run_i(lines, file, output):
    extra = dict(python=None, output=output, evaluated_lines=0)
    def block_fun(lines, start_extra, end_extra, art, head="", tail="", output=None, extra=None):
        outf = output + ("_" + art if art is not None and len(art) > 0 else "") + ".shell"
        lines = full_strip(lines)
        s = "\n".join(lines)
        s.replace("...", "..") # passive-aggressively truncate ... because of #issues.
        lines = textwrap.dedent(s).strip().splitlines()

        if extra['python'] is None:
            an = we.spawn("python", encoding="utf-8", timeout=20)
            an.expect([">>>"])
            extra['python'] = an

        analyzer = extra['python']
        def rsession(analyzer, lines):
            l2 = []
            for i, l in enumerate(lines):
                l2.append(l)
                if l.startswith(" ") and i < len(lines)-1 and not lines[i+1].startswith(" "):
                    if not lines[i+1].strip().startswith("else:") and not lines[i+1].strip().startswith("elif") :
                        l2.append("\n")

            lines = l2
            alines = []
            in_dot_mode = False
            if len(lines[-1]) > 0 and (lines[-1].startswith(" ") or lines[-1].startswith("\t")):
                lines += [""]

            for i, word in enumerate(lines):
                analyzer.sendline(word)
                before = ""
                while True:
                    analyzer.expect_exact([">>>", "..."])
                    before += analyzer.before
                    if analyzer.before.endswith("\n"):
                        break
                    else:
                        before += analyzer.after

                dotmode = analyzer.after == "..."
                if 'dir(s)' in word:
                    pass
                if 'help(s.find)' in word:
                    pass
                if dotmode:
                    alines.append(">>>" + analyzer.before.rstrip() if not in_dot_mode else "..." + analyzer.before.rstrip())
                    in_dot_mode = True
                else:
                    alines.append( ("..." if in_dot_mode else ">>>") + analyzer.before.rstrip())
                    in_dot_mode = False
            return alines

        for l in (head[extra['evaluated_lines']:] + ["\n"]):
            analyzer.sendline(l)
            analyzer.expect_exact([">>>", "..."])


        alines = rsession(analyzer, lines)
        extra['evaluated_lines'] += len(head) + len(lines)
        lines = alines
        return lines, [outf, lines]
    try:
        a,b,c,_ = block_process(lines, tag="#!i", block_fun=functools.partial(block_fun, output=output, extra=extra))
        if extra['python'] is not None:
            extra['python'].close()

        if len(c)>0:
            kvs= { v[0] for v in c}
            for outf in kvs:
                out = "\n".join( ["\n".join(v[1]) for v in c if v[0] == outf] )
                out = out.replace("\r", "")

                with open(outf, 'w') as f:
                    f.write(out)

    except Exception as e:
        print("lines are")
        print("\n".join(lines))
        print("Bad thing in #!i command in file", file)
        raise e
    return lines


from collections import defaultdict
import os
from snipper.block_parsing import block_iterate
from snipper.snipper_main import full_strip

def get_s(lines):
    """ Return snips from 'lines' """
    blocks = defaultdict(list)
    for c in block_iterate(lines, "#!s"):
        # c['start_tag_args']
        if not c['start_tag_args'].get('keeptags', False):
            c['block'] = full_strip(c['block'])
        else:
            # In this case the #! tags are kept in.
            pass
            # print("keepting tags.")
        blocks[c['name']].append(c)

    output = {}
    for name, co in blocks.items():
        slines = [l for c in co for l in c['block']]
        # full_strip("")
        # c['block']['args']
        # slines = slines[ 23]
        # co.
        output[name] = slines
    return output

# def _s_block_process():
#
#     pass

def save_s(lines, output_dir, file_path): # save file snips to disk
    content = get_s(lines)
    # Only make output dir if needed.
    if len(content) > 0 and not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    for name, ll in content.items():
        if file_path is not None:
            file_path = file_path.replace("\\", "/")
            ll = [f"# {file_path}"] + ll
        out = "\n".join(ll)

        with open(output_dir + "/" + os.path.basename(file_path)[:-3] + ("_" + name if len(name) > 0 else name) + ".py", 'w') as f:
            f.write(out)


s1 = """
L1
L2 #!s=a
L3 #!s=b
L4
L5 
L6 
L7 #!s=a
L8 
L9 #!s=b
went
"""
if __name__ == "__main__":
    # for c in block_iterate(s1.splitlines(), tag="#!s"):
    #     print(c['block'])
    output = get_s(s1.splitlines())
    for k, v in output.items():
        print("name:", k)
        print("\n".join(v))
        print("="*10)

